/*--------------------------------------------------------
   SampleFilter.c -- Standard Input/Output Filter Program
             UTF-8/UTF-16  Shift-JIS ϊ
   Copyright (c) 2005
        Hiroto Shigeta, All rights reserved.
  --------------------------------------------------------*/
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <locale.h>
#include <fcntl.h>
#include <io.h>

#define IS_UTF8_2(ch1,ch2)		(ch1>=0xC2&&ch1<=0xDF&&ch2>=0x80&&ch2<=0xBF)
#define IS_UTF8_3(ch1,ch2,ch3)	((ch1==0xE0&&ch2>=0xA0&&ch2<=0xBF&&ch3>=0x80&&ch3<=0xBF)||\
								 (ch1>=0xE1&&ch1<=0xEC&&ch2>=0x80&&ch2<=0xBF&&ch3>=0x80&&ch3<=0xBF)||\
								 (ch1==0xED&&ch2>=0x80&&ch2<=0x9F&&ch3>=0x80&&ch3<=0xBF)||\
								 (ch1>=0xEE&&ch1<=0xEF&&ch2>=0x80&&ch2<=0xBF&&ch3>=0x80&&ch3<=0xBF))

#define RET_ERROR     -1	/* ϊG[ */
#define RET_SHIFT_JIS  0	/* Shift-JIS(ANSI) */
#define RET_UTF8       1	/* UTF-8 VOj`Ȃ */
#define RET_UTF8S      2	/* UTF-8 VOj`t */
#define RET_UTF16      3	/* UTF-16 gGfBA(BOMt) */
#define RET_UTF16B     4	/* UTF-16 rbOGfBA(BOMt) */

static unsigned char byBuffer[BUFSIZ];	/* ʂɑ傫obt@ */
static int           iCur = 0;			/* ݈ʒu */
static int           iEnd = 0;			/* I_ */
static BOOL          bBuffering = TRUE;	/* obt@Os */

static int getone(void)
{
	int ch;

	if (iCur < iEnd) {
		return (int)byBuffer[iCur++];
	}
	ch = getchar();
	if (ch != EOF) {
		if (bBuffering) {
			byBuffer[iEnd] = ch;
			iCur = ++iEnd;
		} else {
			iCur = iEnd = 0;
		}
	}
	return ch;
}

static void pushone(unsigned char ch)
{
	bBuffering = TRUE;
	if (iCur == iEnd) {
		byBuffer[0] = ch;
		iCur = iEnd = 1;
	}
}

static void restore(int count)
{
	bBuffering = FALSE;
	iCur = max(0, iCur - count);
}

static void errorchar(BOOL bBig, BOOL bUtf8)
{
	if (bUtf8) {
		putchar('?');
	} else if (bBig) {
		putchar('\0');
		putchar('?');
	} else {
		putchar('?');
		putchar('\0');
	}
}

static int sjis2unicode(BOOL bBig, BOOL bUtf8)
{
	int           ch;
	unsigned char in[2];
	unsigned char out[3];
	size_t        count;
	int           ret;
	wchar_t       wch;
	BOOL          bError = FALSE;

	while ((ch = getchar()) != EOF) {
		in[0] = ch;
		if (IsDBCSLeadByte(in[0])) {
			if ((ch = getchar()) == EOF) {
				OutputDebugString("DBCSoCgI\n");
				errorchar(bBig, bUtf8);
				bError = TRUE;
				break;
			}
			in[1] = ch;
			count = 2;
		} else {
			count = 1;
		}
		ret = mbtowc((wchar_t *)out, in, count);
		if (ret < 1) {
			OutputDebugString("Shift-JISUNICODEϊG[\n");
			errorchar(bBig, bUtf8);
			bError = TRUE;
			continue;
		}
		if (bUtf8) {
			wch = *(wchar_t *)out;
			if (wch <= 0x007F) {
				putchar(out[0]);
			} else if (wch <= 0x07FF) {
				out[0] = (unsigned char)(wch >> 6) | 0xC0;
				out[1] = (unsigned char)(wch & 0x003F) | 0x80;
				putchar(out[0]);
				putchar(out[1]);
			} else {
				out[0] = (unsigned char)(wch >> 12) | 0xE0;
				out[1] = (unsigned char)(wch >> 6 & 0x003F) | 0x80;
				out[2] = (unsigned char)(wch & 0x003F) | 0x80;
				putchar(out[0]);
				putchar(out[1]);
				putchar(out[2]);
			}
		} else if (bBig) {
			putchar(out[1]);
			putchar(out[0]);
		} else {
			putchar(out[0]);
			putchar(out[1]);
		}
	}
	return bError ? RET_ERROR : RET_SHIFT_JIS;
}

static int utf16tosjis(BOOL bBig)
{
	int           ch;
	unsigned char in[2];
	unsigned char out[2];
	int           ret;
	BOOL          bError = FALSE;

	while ((ch = getchar()) != EOF) {
		if (bBig) in[1] = ch; else in[0] = ch;
		if ((ch = getchar()) == EOF) {
			OutputDebugString("UTF16oCgI\n");
			putchar('?');
			bError = TRUE;
			break;
		}
		if (bBig) in[0] = ch; else in[1] = ch;
		ret = wctomb(out, *(wchar_t *)in);
		if (ret < 1) {
			OutputDebugString("UTF16Shift-JISϊG[\n");
			putchar('?');
			bError = TRUE;
			continue;
		}
		putchar(out[0]);
		if (ret > 1) {
			putchar(out[1]);
		}
	}
	return bError ? RET_ERROR : bBig ? RET_UTF16B : RET_UTF16;
}

static int utf8tosjis(void)
{
	int           ch;
	unsigned char ch1, ch2, ch3;
	wchar_t       wch;
	unsigned char out[2];
	int           ret;
	BOOL          bError = FALSE;

	while ((ch = getone()) != EOF) {
		ch1 = ch;
		if (ch1 <= 0x7F) {
			putchar(ch1);
			continue;
		}
		ch = getone();
		if (ch == EOF) {
			OutputDebugString("UTF8 1oCg]\n");
			putchar('?');
			bError = TRUE;
			break;
		}
		ch2 = ch;
		if (IS_UTF8_2(ch1, ch2)) {
			wch = ch1 & 0x1F;
			wch = wch << 6;
			wch |= ch2 & 0x3F;
			ret = wctomb(out, wch);
			if (ret < 1) {
				OutputDebugString("UTF8(2oCg)Shift-JISϊG[\n");
				putchar('?');
				bError = TRUE;
				continue;
			}
			putchar(out[0]);
			if (ret > 1) {
				putchar(out[1]);
			}
			continue;
		}
		ch = getone();
		if (ch == EOF) {
			OutputDebugString("UTF8 2oCg]\n");
			putchar('?');
			bError = TRUE;
			break;
		}
		ch3 = ch;
		wch = ch1 & 0x0F;
		wch = wch << 6;
		wch |= ch2 & 0x3F;
		wch = wch << 6;
		wch |= ch3 & 0x3F;
		ret = wctomb(out, wch);
		if (ret < 1) {
			OutputDebugString("UTF8(3oCg)Shift-JISϊG[\n");
			putchar('?');
			bError = TRUE;
			continue;
		}
		putchar(out[0]);
		if (ret > 1) {
			putchar(out[1]);
		}
	}
	return bError ? RET_ERROR : RET_UTF8;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR pszCmdLine, int iCmdShow)
{
	int           ch;
	unsigned char ch1, ch2, ch3;

	/* ܂Ȃ */
	_setmode(_fileno(stdin), _O_BINARY);
	_setmode(_fileno(stdout), _O_BINARY);
	setlocale(LC_ALL, "japanese");
	/* R}hC`FbN */
	if (strnicmp(pszCmdLine, "/oUTF16B", 8) == 0) {
		putchar(0xFE);
		putchar(0xFF);
		return sjis2unicode(TRUE, FALSE);
	} else if (strnicmp(pszCmdLine, "/oUTF16", 7) == 0) {
		putchar(0xFF);
		putchar(0xFE);
		return sjis2unicode(FALSE, FALSE);
	} else if (strnicmp(pszCmdLine, "/oUTF8S", 7) == 0) {
		putchar(0xEF);
		putchar(0xBB);
		putchar(0xBF);
		return sjis2unicode(FALSE, TRUE);
	} else if (strnicmp(pszCmdLine, "/oUTF8", 6) == 0) {
		return sjis2unicode(FALSE, TRUE);
	} else if (strnicmp(pszCmdLine, "/o", 2) == 0) {
		while ((ch = getchar()) != EOF) {
			putchar(ch);
		}
		return RET_SHIFT_JIS;
	}
	/* F */
	ch = getone();
	if (ch == EOF) {
		/* t@C */
		return RET_SHIFT_JIS;
	}
	ch1 = ch;
	ch = getone();
	if (ch == EOF) {
		/* 1bytẽt@C */
		putchar(ch1);
		return RET_SHIFT_JIS;
	}
	ch2 = ch;
	if (ch1 == 0xFF && ch2 == 0xFE) {
		/* UTF-16 gGfBA */
		return utf16tosjis(FALSE);
	} else if (ch1 == 0xFE && ch2 == 0xFF) {
		/* UTF-16 rbOGfBA */
		return utf16tosjis(TRUE);
	}
	ch = getone();
	if (ch == EOF) {
		/* 2bytẽt@C */
		if (IS_UTF8_2(ch1, ch2)) {
			/* UTF-8 VOj`Ȃ */
			restore(2);
			return utf8tosjis();
		} else {
			putchar(ch1);
			putchar(ch2);
			return RET_SHIFT_JIS;
		}
	}
	ch3 = ch;
	if (ch1 == 0xEF && ch2 == 0xBB && ch3 == 0xBF) {
		/* UTF-8 VOj`t */
		restore(0);
		return utf8tosjis() == RET_ERROR ? RET_ERROR : RET_UTF8S;
	}
	restore(3);
	while ((unsigned char)(ch = getone()) <= 0x7F) {
		putchar(ch);
	}
	if (ch == EOF) {
		return RET_SHIFT_JIS;
	}
	ch1 = ch;
	pushone(ch1);
	ch = getone();
	if (ch == EOF) {
		putchar(ch1);
		return RET_SHIFT_JIS;
	}
	ch2 = ch;
	if (IS_UTF8_2(ch1, ch2)) {
		/* UTF-8 VOj`Ȃ */
		restore(2);
		return utf8tosjis();
	}
	ch = getone();
	if (ch == EOF) {
		putchar(ch1);
		putchar(ch2);
		return RET_SHIFT_JIS;
	}
	ch3 = ch;
	if (IS_UTF8_3(ch1, ch2, ch3)) {
		/* UTF-8 VOj`Ȃ */
		restore(3);
		return utf8tosjis();
	}
	putchar(ch1);
	putchar(ch2);
	putchar(ch3);
	while ((ch = getchar()) != EOF) {
		putchar(ch);
	}
	return RET_SHIFT_JIS;
}
